from .base import BaseCaptcha
from .enums import KeyCaptchaEnm


class BaseKeyCaptcha(BaseCaptcha):
    def __init__(
        self,
        pageurl: str,
        s_s_c_user_id: str,
        s_s_c_session_id: str,
        s_s_c_web_server_sign: str,
        s_s_c_web_server_sign2: str,
        method: str = KeyCaptchaEnm.KEYCAPTCHA.value,
        *args,
        **kwargs,
    ):
        super().__init__(method=method, *args, **kwargs)

        self.post_payload.update(
            {
                "pageurl": pageurl,
                "s_s_c_user_id": s_s_c_user_id,
                "s_s_c_session_id": s_s_c_session_id,
                "s_s_c_web_server_sign": s_s_c_web_server_sign,
                "s_s_c_web_server_sign2": s_s_c_web_server_sign2,
            }
        )

        # check user params
        if method not in KeyCaptchaEnm.list_values():
            raise ValueError(f"Invalid method parameter set, available - {KeyCaptchaEnm.list_values()}")


class KeyCaptcha(BaseKeyCaptcha):
    """
    The class is used to work with KeyCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_keycaptcha
    """

    def captcha_handler(self, **kwargs):
        """
        The method is responsible for sending data to the server to solve the captcha
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """

        return self._processing_response(**kwargs)


class aioKeyCaptcha(BaseKeyCaptcha):

    """
    The class is used to async work with KeyCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_keycaptcha
    """

    async def captcha_handler(self):
        """
        The method is responsible for sending data to the server to solve the captcha
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        return await self._aio_processing_response()



'''
class WebSocketRuCaptcha(BaseCaptcha):
    def __init__(self, allSessions: bool = False, suppressSuccess: bool = True):
        """
        Method setup WebSocket connection data
        :param allSessions: `True` if u need send the results of solving captchas to all open sessions.
                            `False` if u will wait captcha result in current session.
        :param suppressSuccess: `False` if u need intermediate info about your task in RuCaptcha system.
                                `True` if u need final answer without any addition info.
        More info - https://wsrucaptcha.docs.apiary.io/#reference/0
        """
        self.sock: WebSocketClientProtocol = None
        self.result = SocketResponse()
        self.auth_result = SocketResponse()

        self.allSessions = allSessions or False
        self.suppressSuccess = suppressSuccess or True
        # RuCaptcha auth params check
        if not self.allSessions and not self.suppressSuccess:
            raise ValueError(
                f"U set allSessions to `{self.allSessions}` and suppressSuccess to `{self.suppressSuccess}`.\n"
                f"U will not able to get the final result of captcha solving, cause u close the current socket session "
                f"but wait for the result ONLY in the current session(already closed).\n"
                f"Try other param variants:\n"
                f"\tlike default False/True(u will get an answer in this session for this captcha)\n"
                f"\tor True/False(u will get info about the current captcha task and then get captcha solution in all "
                f"opened sessions)"
            )

        self.URL = "wss://s.2captcha.com/"
        # WebSocket SSL setup
        self.ssl_context = ssl.create_default_context()
        self.ssl_context.check_hostname = False
        self.ssl_context.verify_mode = ssl.CERT_NONE

    async def __socket_session_close(self):
        """
        Method close exist socket session
        """
        if self.sock:
            await self.sock.close()

    async def __socket_session(self):
        """
        Create new socket session
        """
        self.sock = await websockets.connect(self.URL, ssl=self.ssl_context)

    async def __socket_session_recreate(self):
        """
        If socket session not exist - create it
        """
        # if socket not exist
        if not self.sock:
            await self.__socket_session()
        # if socket exist but closed
        elif self.sock.closed:
            await self.__socket_session_close()
            await self.__socket_session()

    async def __auth(self):
        """
        Method setup connection with RuCaptcha and auth user
        """
        await self.__socket_session_recreate()

        auth_data = SockAuthSer(
            **{
                "key": self.rucaptcha_key,
                "options": {"allSessions": self.allSessions, "suppressSuccess": self.suppressSuccess},
            }
        )
        await self.sock.send(auth_data.json())
        self.auth_result = self.auth_result.parse_raw(await self.sock.recv())

    @retry(wait=wait_fixed(5), stop=stop_after_attempt(3), after=after_log(logging, logging.ERROR), reraise=True)
    async def get_request(self) -> dict:
        """
        Method send GET request and write result to response model
        :return: server response info, serialized to dict without NONE values
        """
        await self.__socket_session_recreate()

        response = await self.sock.recv()
        await self.__socket_session_close()

        return self.result.parse_raw(response).dict(exclude_none=True)

    @retry(wait=wait_fixed(5), stop=stop_after_attempt(3), after=after_log(logging, logging.ERROR), reraise=True)
    async def send_request(self, payload: str) -> dict:
        """
        Method send request to server and wait response
        :param payload: JSON payload with data
        :return: Server response dict
        """
        await self.__auth()
        # check if auth is success
        if self.auth_result.success:
            await self.sock.send(payload)
            return await self.get_request()
        else:
            await self.__socket_session_close()
            return self.auth_result.dict(exclude_none=True)
'''



import time
import asyncio

import aiohttp
import requests

from .config import connect_generator
from .serializer import ResponseSer, ServiceGetResponseSer


def result_processing(captcha_response: ServiceGetResponseSer, result: ResponseSer) -> dict:
    """
    Function processing service response status values
    """

    # on error during solving
    if captcha_response.status == 0:
        result.error = True
        result.errorBody = captcha_response.request

    # if solving is success
    elif captcha_response.status == 1:
        result.captchaSolve = captcha_response.request

        # if this is ReCaptcha v3 then we get it from the server
        if captcha_response.user_check and captcha_response.user_score:
            result = result.dict()
            result.update(
                {
                    "user_check": captcha_response.user_check,
                    "user_score": captcha_response.user_score,
                }
            )
            return result

    return result.dict()


def get_sync_result(get_payload: dict, sleep_time: int, url_response: str, result: ResponseSer) -> dict:
    """
    Function periodically send the SYNC request to service and wait for captcha solving result
    """
    # generator for repeated attempts to connect to the server
    connect_gen = connect_generator()
    while True:
        try:
            # send a request for the result of solving the captcha
            captcha_response = ServiceGetResponseSer(**requests.get(url_response, params=get_payload).json())
            # if the captcha has not been resolved yet, wait
            if captcha_response.request == "CAPCHA_NOT_READY":
                time.sleep(sleep_time)
            else:
                return result_processing(captcha_response, result)

        except Exception as error:
            if next(connect_gen) < 4:
                time.sleep(2)
            else:
                result.error = True
                result.errorBody = error
                return result.dict()


async def get_async_result(get_payload: dict, sleep_time: int, url_response: str, result: ResponseSer) -> dict:
    """
    Function periodically send the ASYNC request to service and wait for captcha solving result
    """
    # generator for repeated attempts to connect to the server
    connect_gen = connect_generator()
    async with aiohttp.ClientSession() as session:
        while True:
            try:
                # send a request for the result of solving the captcha
                async with session.get(url_response, params=get_payload, raise_for_status=True) as resp:
                    captcha_response = await resp.json()
                    captcha_response = ServiceGetResponseSer(**captcha_response)

                    # if the captcha has not been resolved yet, wait
                    if captcha_response.request == "CAPCHA_NOT_READY":
                        await asyncio.sleep(sleep_time)

                    else:
                        return result_processing(captcha_response, result)

            except Exception as error:
                if next(connect_gen) < 4:
                    await asyncio.sleep(2)
                else:
                    result.error = True
                    result.errorBody = error
                    return result.dict()



import time
import asyncio

import aiohttp
import requests
from requests.adapters import HTTPAdapter

from . import enums
from .config import RETRIES, ASYNC_RETRIES
from .serializer import ResponseSer, GetRequestSer, PostRequestSer, CaptchaOptionsSer, ServicePostResponseSer
from .result_handler import get_sync_result, get_async_result


class BaseCaptcha:
    def __init__(
        self,
        rucaptcha_key: str,
        method: str = "",
        action: str = "get",
        sleep_time: int = 15,
        service_type: str = enums.ServicesEnm.TWOCAPTCHA.value,
        **kwargs,
    ):
        """
        :param rucaptcha_key: User API key
        :param method: Captcha type
        :param action: Server action
        :param sleep_time: Time to wait for captcha solution
        :param service_type: URL with which the program will work, "2captcha" option is possible (standard)
                              and "rucaptcha"
        :param kwargs: Designed to pass OPTIONAL parameters to the payload for a request to RuCaptcha
        """
        # assign args to validator
        self.params = CaptchaOptionsSer(**locals())

        # prepare POST payload
        self.post_payload = PostRequestSer(key=self.params.rucaptcha_key, method=method).dict(by_alias=True)
        # prepare GET payload
        self.get_payload = GetRequestSer(key=self.params.rucaptcha_key, action=action).dict(
            by_alias=True, exclude_none=True
        )
        # prepare result payload
        self.result = ResponseSer()

        # If more parameters are passed, add them to post_payload
        if kwargs:
            for key in kwargs:
                self.post_payload.update({key: kwargs[key]})

        # prepare session
        self.session = requests.Session()
        self.session.mount("http://", HTTPAdapter(max_retries=RETRIES))
        self.session.mount("https://", HTTPAdapter(max_retries=RETRIES))

    def _processing_response(self, **kwargs: dict) -> dict:
        """
        Method processing captcha solving task creation result
        :param kwargs: additional params for Requests library
        """
        try:
            response = ServicePostResponseSer(
                **self.session.post(self.params.url_request, data=self.post_payload, **kwargs).json()
            )
            # check response status
            if response.status == 1:
                self.result.taskId = response.request
            else:
                self.result.error = True
                self.result.errorBody = response.request
        except Exception as error:
            self.result.error = True
            self.result.errorBody = error

        # check for errors while make request to server
        if self.result.error:
            return self.result.dict(exclude_none=True)

        # if all is ok - send captcha to service and wait solution
        # update payload - add captcha taskId
        self.get_payload.update({"id": self.result.taskId})

        # wait captcha solving
        time.sleep(self.params.sleep_time)
        return get_sync_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )

    def url_open(self, url: str, **kwargs):
        """
        Method open links
        """
        return self.session.get(url=url, **kwargs)

    async def aio_url_read(self, url: str, **kwargs) -> bytes:
        """
        Async method read bytes from link
        """
        async with aiohttp.ClientSession() as session:
            async for attempt in ASYNC_RETRIES:
                with attempt:
                    async with session.get(url=url, **kwargs) as resp:
                        return await resp.content.read()

    async def _aio_processing_response(self) -> dict:
        """
        Method processing async captcha solving task creation result
        """
        try:
            # make async or sync request
            response = await self.__aio_make_post_request()
            # check response status
            if response.status == 1:
                self.result.taskId = response.request
            else:
                self.result.error = True
                self.result.errorBody = response.request
        except Exception as error:
            self.result.error = True
            self.result.errorBody = error

        # check for errors while make request to server
        if self.result.error:
            return self.result.dict(exclude_none=True)

        # if all is ok - send captcha to service and wait solution
        # update payload - add captcha taskId
        self.get_payload.update({"id": self.result.taskId})

        # wait captcha solving
        await asyncio.sleep(self.params.sleep_time)
        return await get_async_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )

    async def __aio_make_post_request(self) -> ServicePostResponseSer:
        async with aiohttp.ClientSession() as session:
            async for attempt in ASYNC_RETRIES:
                with attempt:
                    async with session.post(
                        self.params.url_request, data=self.post_payload, raise_for_status=True
                    ) as resp:
                        response_json = await resp.json()
                        return ServicePostResponseSer(**response_json)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return False
        return True

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        if exc_type:
            return False
        return True



from .base import BaseCaptcha


class BaseTextCaptcha(BaseCaptcha):
    def __init__(
        self,
        language: int = 0,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        self.post_payload.update({"language": language})


class TextCaptcha(BaseTextCaptcha):
    """
    The class is used to work with TextCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_text_captcha
    """

    def captcha_handler(self, textcaptcha: str, **kwargs):
        """
        The method is responsible for sending data to the server to solve the captcha
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        self.post_payload.update({"textcaptcha": textcaptcha})

        return self._processing_response(**kwargs)


class aioTextCaptcha(BaseTextCaptcha):
    """
    The class is used to async work with TextCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_text_captcha
    """

    async def captcha_handler(self, textcaptcha: str):
        """
        The method is responsible for sending data to the server to solve the captcha
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        self.post_payload.update({"textcaptcha": textcaptcha})
        return await self._aio_processing_response()


'''class sockTextCaptcha(WebSocketRuCaptcha):
    def __init__(self, rucaptcha_key: str):
        """
        The asynchronous WebSocket module is responsible for text captcha solving.
        :param rucaptcha_key: Key from RuCaptcha
        """
        super().__init__()
        self.rucaptcha_key = rucaptcha_key

    async def captcha_handler(self, captcha_text: str, **kwargs) -> dict:
        """
        The asynchronous WebSocket method return account balance.
        More info - https://wsrucaptcha.docs.apiary.io/#reference/text-captcha
        :param captcha_text: Text captcha task string
        :param kwargs: Options variables
        :return: Server response dict
        """
        text_captcha_payload = TextCaptchaSer(
            **{
                "method": "text",
                "requestId": str(uuid4()),
                "text": captcha_text,
                "options": CaptchaOptionsSer(**kwargs),
            }
        )
        return await self.send_request(text_captcha_payload.dict())
'''



import os
import uuid
import base64
import shutil

from .base import BaseCaptcha
from .enums import SaveFormatsEnm, ImageCaptchaEnm


class BaseImageCaptcha(BaseCaptcha):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(method=ImageCaptchaEnm.BASE64.value, *args, **kwargs)

    def _image_const_saver(self, content: bytes, img_path: str):
        """
        Method create and save file in folder
        """
        # generate image name
        self._image_name = f"im-{uuid.uuid4()}.png"

        # save image to folder
        with open(os.path.join(img_path, self._image_name), "wb") as out_image:
            out_image.write(content)

    @staticmethod
    def _local_image_captcha(captcha_file: str):
        """
        Method get local image, read it and prepare for sending to Captcha solving service
        """
        with open(captcha_file, "rb") as file:
            return file.read()


class ImageCaptcha(BaseImageCaptcha):
    """
    The class is used to work with ImageCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_normal_captcha
    """

    def __init__(
        self,
        save_format: str = SaveFormatsEnm.TEMP.value,
        img_clearing: bool = True,
        img_path: str = "PythonRuCaptchaImages",
        *args,
        **kwargs,
    ):
        """
        Initializing the necessary variables, creating a folder for images and cache
        After completion of work - temporary files and folders will be deleted
        :param save_format: The format in which the image will be saved, or as a temporary file - 'temp',
                             or as a regular image to a folder created by the library - 'const'.
        :param img_clearing: True - delete file after solution, False - don't delete file after solution;
        :param img_path: Folder to save captcha images;
        """
        super().__init__(*args, **kwargs)
        self.save_format = save_format
        self.img_clearing = img_clearing
        self.img_path = img_path

    def captcha_handler(
        self,
        captcha_link: str = None,
        captcha_file: str = None,
        captcha_base64: bytes = None,
        **kwargs,
    ) -> dict:
        """
        The method is responsible for sending data to the server to solve the captcha
        :param captcha_link: Captcha image URL
        :param captcha_file: Captcha image file path
        :param captcha_base64: Captcha image BASE64 info
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        # if a local file link is passed
        if captcha_file:
            self.post_payload.update(
                {"body": base64.b64encode(self._local_image_captcha(captcha_file)).decode("utf-8")}
            )
        # if the file is transferred in base64 encoding
        elif captcha_base64:
            self.post_payload.update({"body": base64.b64encode(captcha_base64).decode("utf-8")})
        # if a URL is passed
        elif captcha_link:
            try:
                content = self.url_open(url=captcha_link, **kwargs).content
            except Exception as error:
                self.result.error = True
                self.result.errorBody = error
                return self.result.dict(exclude_none=True)

            # according to the value of the passed parameter, select the function to save the image
            if self.save_format == SaveFormatsEnm.CONST.value:
                self._image_const_saver(content, self.img_path)
            self.post_payload.update({"body": base64.b64encode(content).decode("utf-8")})

        else:
            # if none of the parameters are passed
            self.result.error = True
            self.result.errorBody = "You did not send any file, local link or URL."
            return self.result.dict(exclude_none=True)

        return self._processing_response(**kwargs)

    def __del__(self):
        if self.save_format == SaveFormatsEnm.CONST.value and self.img_clearing:
            shutil.rmtree(self.img_path)


class aioImageCaptcha(BaseImageCaptcha):
    """
    The class is used to async work with ImageCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_normal_captcha
    """

    def __init__(
        self,
        save_format: str = SaveFormatsEnm.TEMP.value,
        img_clearing: bool = True,
        img_path: str = "PythonRuCaptchaImages",
        *args,
        **kwargs,
    ):
        """
        Initializing the necessary variables, creating a folder for images and cache
        After completion of work - temporary files and folders will be deleted
        :param save_format: The format in which the image will be saved, or as a temporary file - 'temp',
                             or as a regular image to a folder created by the library - 'const'.
        :param img_clearing: True - delete file after solution, False - don't delete file after solution;
        :param img_path: Folder to save captcha images;
        """
        super().__init__(*args, **kwargs)
        self.save_format = save_format
        self.img_clearing = img_clearing
        self.img_path = img_path

    async def captcha_handler(
        self,
        captcha_link: str = None,
        captcha_file: str = None,
        captcha_base64: bytes = None,
        **kwargs,
    ) -> dict:
        """
        The method is responsible for sending data to the server to solve the captcha
        :param captcha_link: Captcha image URL
        :param captcha_file: Captcha image file
        :param captcha_base64: Captcha image BASE64 info
        :param kwargs: Additional parameters for the `aiohttp` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        # if a local file link is passed
        if captcha_file:
            self.post_payload.update(
                {"body": base64.b64encode(self._local_image_captcha(captcha_file)).decode("utf-8")}
            )
        # if the file is transferred in base64 encoding
        elif captcha_base64:
            self.post_payload.update({"body": base64.b64encode(captcha_base64).decode("utf-8")})
        # if a URL is passed
        elif captcha_link:
            try:
                content = await self.aio_url_read(url=captcha_link, **kwargs)
            except Exception as error:
                self.result.error = True
                self.result.errorBody = error
                return self.result.dict(exclude_none=True)

            # according to the value of the passed parameter, select the function to save the image
            if self.save_format == SaveFormatsEnm.CONST.value:
                self._image_const_saver(content, self.img_path)
            self.post_payload.update({"body": base64.b64encode(content).decode("utf-8")})

        else:
            # if none of the parameters are passed
            self.result.error = True
            self.result.errorBody = "You did not send any file, local link or URL."
            return self.result.dict(exclude_none=True)

        return await self._aio_processing_response()

    def __del__(self):
        if self.save_format == SaveFormatsEnm.CONST.value and self.img_clearing:
            shutil.rmtree(self.img_path)


'''
class sockNormalCaptcha(WebSocketRuCaptcha):
    """
    Class for ImageCaptcha
    """

    def __init__(self, rucaptcha_key: str, allSessions: bool = None, suppressSuccess: bool = None):
        """
        Method setup WebSocket connection data
        """
        super().__init__(allSessions, suppressSuccess)
        self.rucaptcha_key = rucaptcha_key

    async def captcha_handler(self, captcha_image_base64: str, **kwargs) -> dict:
        """
        The asynchronous WebSocket method return account balance.
        More info - https://wsrucaptcha.docs.apiary.io/#reference/text-captcha
        :param captcha_image_base64: Image captcha base64 data in string format (decoded in utf-8)
        :param kwargs: Options variables
        :return: Server response dict
        """
        normal_captcha_payload = NormalCaptchaSocketSer(
            **{
                "method": "normal",
                "requestId": str(uuid4()),
                "body": captcha_image_base64,
                "options": CaptchaOptionsSocketSer(**kwargs),
            }
        )

        return await self.send_request(normal_captcha_payload.dict())
'''



import base64

import aiohttp

from .base import BaseCaptcha
from .enums import RotateCaptchaEnm


class BaseRotateCaptcha(BaseCaptcha):
    def __init__(self, method: str = RotateCaptchaEnm.ROTATECAPTCHA.value, *args, **kwargs):
        super().__init__(method=method, *args, **kwargs)

        # check user params
        if method not in RotateCaptchaEnm.list_values():
            raise ValueError(f"Invalid method parameter set, available - {RotateCaptchaEnm.list_values()}")


class RotateCaptcha(BaseRotateCaptcha):
    """
    The class is used to work with RotateCaptcha
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_rotatecaptcha
    """

    def captcha_handler(self, captcha_link: str, **kwargs):
        """
        The method is responsible for sending data to the server to solve the captcha
        :param captcha_link: link or path to captcha file
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        # if user send link - download file
        if "http" in captcha_link:
            content: bytes = self.url_open(url=captcha_link, **kwargs).content
        # is user send file path - read ir
        else:
            with open(captcha_link, "rb") as captcha_image:
                content: bytes = captcha_image.read()
        # decode file data in base64 and prepare payload
        self.post_payload.update({"body": base64.b64encode(content).decode("utf-8")})

        return self._processing_response(**kwargs)


class aioRotateCaptcha(BaseRotateCaptcha):
    """
    The class is used to async work with RotateCaptcha
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_rotatecaptcha
    """

    async def captcha_handler(self, captcha_link: str, **kwargs):
        """
        The method is responsible for sending data to the server to solve the captcha
        :param captcha_link: link or path to captcha file
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        # if user send link - download file
        if "http" in captcha_link:
            async with aiohttp.ClientSession() as session:
                async with session.get(url=captcha_link) as resp:
                    content: bytes = await resp.content.read()
        # if user send file path - read it
        else:
            with open(captcha_link, "rb") as captcha_image:
                content: bytes = captcha_image.read()
        # decode file data in base64 and prepare payload
        self.post_payload.update({"body": base64.b64encode(content).decode("utf-8")})

        return await self._aio_processing_response()



from .base import BaseCaptcha
from .enums import TikTokCaptchaEnm


class BaseTikTokCaptcha(BaseCaptcha):
    def __init__(
        self,
        pageurl: str,
        cookies: str,
        aid: str,
        host: str,
        method: str = TikTokCaptchaEnm.TIKTOK.value,
        *args,
        **kwargs,
    ):
        super().__init__(method=method, *args, **kwargs)

        self.post_payload.update(
            {
                "pageurl": pageurl,
                "cookies": cookies,
                "aid": aid,
                "host": host,
            }
        )

        # check user params
        if method not in TikTokCaptchaEnm.list_values():
            raise ValueError(f"Invalid method parameter set, available - {TikTokCaptchaEnm.list_values()}")


class TikTokCaptcha(BaseTikTokCaptcha):
    """
    The class is used to work with TikTokCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_tiktok
    """

    def captcha_handler(self, **kwargs):
        """
        The method is responsible for sending data to the server to solve the captcha
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """

        return self._processing_response(**kwargs)


class aioTikTokCaptcha(BaseTikTokCaptcha):
    """
    The class is used to async work with TikTokCaptcha.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_tiktok
    """

    async def captcha_handler(self):
        """
        The method is responsible for sending data to the server to solve the captcha
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        return await self._aio_processing_response()



from .base import BaseCaptcha
from .enums import FunCaptchaEnm


class BaseFunCaptcha(BaseCaptcha):
    def __init__(self, pageurl: str, publickey: str, method: str = FunCaptchaEnm.FUNCAPTCHA.value, *args, **kwargs):
        super().__init__(method=method, *args, **kwargs)

        self.post_payload.update({"publickey": publickey, "pageurl": pageurl})

        # check user params
        if method not in FunCaptchaEnm.list_values():
            raise ValueError(f"Invalid method parameter set, available - {FunCaptchaEnm.list_values()}")


class FunCaptcha(BaseFunCaptcha):
    """
    The class is used to work with FunCaptcha.
    Capy is a captcha in the form of a puzzle
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_funcaptcha_new
    """

    def captcha_handler(self, **kwargs):
        """
        The method is responsible for sending data to the server to solve the captcha
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        return self._processing_response(**kwargs)


class aioFunCaptcha(BaseFunCaptcha):
    """
    The class is used to work with FunCaptcha.
    Capy is a captcha in the form of a puzzle
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_funcaptcha_new
    """

    async def captcha_handler(self):
        """
        The method is responsible for sending data to the server to solve the captcha
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        return await self._aio_processing_response()



from .base import BaseCaptcha
from .enums import CaptchaControlEnm
from .result_handler import get_sync_result, get_async_result


class BaseCaptchaControl(BaseCaptcha):
    def __init__(
        self,
        action: str,
        *args,
        **kwargs,
    ):
        super().__init__(action=action, *args, **kwargs)

        # check user params
        if action not in CaptchaControlEnm.list_values():
            raise ValueError(f"Invalid method parameter set, available - {CaptchaControlEnm.list_values()}")


class CaptchaControl(BaseCaptchaControl):
    """
    The class is used to work with RuCaptcha control methods.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#manage_pingback
        https://rucaptcha.com/api-rucaptcha#complain
        https://rucaptcha.com/api-rucaptcha#additional
    """

    def domain_control(self, addr: str) -> dict:
        """
        Callback domains control
            https://rucaptcha.com/api-rucaptcha#manage_pingback
        :param addr: Your URL to send a response.
        :return: Dict with result
        """
        self.get_payload.update({"addr": addr})
        return get_sync_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )

    def report(self, id: str) -> dict:
        """
        Callback results report
            https://rucaptcha.com/api-rucaptcha#complain
        :param id: Captcha solving task ID
        :return: Dict with result
        """
        self.get_payload.update({"addr": id})
        return get_sync_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )

    def additional_methods(self, **kwargs) -> dict:
        """
        Callback results report
            https://rucaptcha.com/api-rucaptcha#additional
        :param kwargs: Additional params for method
        :return: Dict with result
        """
        # If more parameters are passed, add them to get_payload
        if kwargs:
            for key in kwargs:
                self.get_payload.update({key: kwargs[key]})
        return get_sync_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )


class aioCaptchaControl(BaseCaptchaControl):
    """
    The class is used to async work with RuCaptcha control methods.
    Solve description:
        https://rucaptcha.com/api-rucaptcha#manage_pingback
        https://rucaptcha.com/api-rucaptcha#complain
        https://rucaptcha.com/api-rucaptcha#additional
    """

    async def domain_control(self, addr: str) -> dict:
        """
        Callback domains control
            https://rucaptcha.com/api-rucaptcha#manage_pingback
        :param addr: Your URL to send a response.
        :return: Dict with result
        """
        self.get_payload.update({"addr": addr})
        return await get_async_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )

    async def report(self, id: str) -> dict:
        """
        Callback results report
            https://rucaptcha.com/api-rucaptcha#complain
        :param id: Captcha solving task ID
        :return: Dict with result
        """
        self.get_payload.update({"addr": id})
        return await get_async_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )

    async def additional_methods(self, **kwargs) -> dict:
        """
        Callback results report
            https://rucaptcha.com/api-rucaptcha#additional
        :param kwargs: Additional params for method
        :return: Dict with result
        """
        # If more parameters are passed, add them to get_payload
        if kwargs:
            for key in kwargs:
                self.get_payload.update({key: kwargs[key]})
        return await get_async_result(
            get_payload=self.get_payload,
            sleep_time=self.params.sleep_time,
            url_response=self.params.url_response,
            result=self.result,
        )


'''
# Async WebSocket method
class sockCaptchaControl(WebSocketRuCaptcha):
    def __init__(self, rucaptcha_key: str, allSessions: bool = None, suppressSuccess: bool = None):
        """
        Method setup WebSocket connection data
        Params description check in parent class
        """
        super().__init__(allSessions, suppressSuccess)
        self.rucaptcha_key = rucaptcha_key

    async def get_balance(self) -> dict:
        """
        The asynchronous WebSocket method return account balance.
        More info - https://wsrucaptcha.docs.apiary.io/#reference/4
        :return: Server response dict
        """
        balance_payload = ControlCaptchaSocketSer(
            **{
                "method": "balance",
            }
        )
        return await self.send_request(balance_payload.json(exclude_none=True))

    async def report(self, success: bool, captchaId: int) -> dict:
        """
        The asynchronous WebSocket method send captcha solving reports (success or fail).
        More info - https://wsrucaptcha.docs.apiary.io/#reference/2
        :param success: Is captcha solved success?
        :param captchaId: Captcha task unique id. For example - 5034284222
        :return: Server response dict
        """
        report_payload = ControlCaptchaSocketSer(
            **{
                "method": "report",
                "success": success,
                "captchaId": captchaId,
            }
        )

        return await self.send_request(report_payload.json(exclude_none=True))
'''



from .base import BaseCaptcha
from .enums import GeetestEnm


class BaseGeeTest(BaseCaptcha):
    def __init__(
        self,
        pageurl: str,
        method: str,
        gt: str = None,
        captcha_id: str = None,
        *args,
        **kwargs,
    ):
        self.method = method
        # check user params
        if self.method not in GeetestEnm.list_values():
            raise ValueError(f"Invalid method parameter set, available - {GeetestEnm.list_values()}")

        super().__init__(method=self.method, *args, **kwargs)
        # insert `gt` param to payload
        self.post_payload.update({"gt": gt, "pageurl": pageurl, "captcha_id": captcha_id})

        if self.method == GeetestEnm.GEETEST_V4.value and captcha_id is None:
            raise ValueError(f"For {self.method} captcha_id is required")
        elif self.method == GeetestEnm.GEETEST.value and gt is None:
            raise ValueError(f"For {self.method} gt is required")


class GeeTest(BaseGeeTest):
    """
    Class for solve Geetest and Geetest v4 captcha
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_geetest
        https://rucaptcha.com/api-rucaptcha#geetest-v4
    """

    def captcha_handler(self, challenge: str = None, **kwargs) -> dict:
        """
        The method is responsible for sending data to the server to solve the captcha
        :param challenge: The value of the challenge parameter found on the site
        :param kwargs: Parameters for the `requests` library
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        if self.method == GeetestEnm.GEETEST.value:
            if challenge is not None:
                self.post_payload.update({"challenge": challenge})
            else:
                raise ValueError(f"For {self.method} challenge is required")

        return self._processing_response(**kwargs)


class aioGeeTest(BaseGeeTest):
    """
    Class for async solve Geetest and Geetest v4 captcha
    Solve description:
        https://rucaptcha.com/api-rucaptcha#solving_geetest
        https://rucaptcha.com/api-rucaptcha#geetest-v4
    """

    async def captcha_handler(self, challenge: str = None) -> dict:
        """
        The method is responsible for sending data to the server to solve the captcha
        :param challenge: The value of the challenge parameter found on the site
        :return: Response to captcha as JSON string with fields:
                 captchaSolve - captcha solution,
                 taskId - finds the ID of the task to solve the captcha,
                 error - False - if everything is fine, True - if there is an error,
                 errorBody - error name
        """
        if self.method == GeetestEnm.GEETEST.value:
            if challenge is not None:
                self.post_payload.update({"challenge": challenge})
            else:
                raise ValueError(f"For {self.method} challenge is required")

        return await self._aio_processing_response()



import io
import os
import sys
from shutil import rmtree

from setuptools import Command, setup
from pkg_resources import parse_requirements

# Package meta-data.
NAME = "python-rucaptcha"
DESCRIPTION = "Python 3.6+ RuCaptcha library with AIO module."
URL = "https://github.com/AndreiDrang/python-rucaptcha"
EMAIL = "drang.andray@gmail.com"
AUTHOR = "AndreiDrang, redV0ID"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "4.2"
with open("requirements.txt", "rt") as requirements_txt:
    REQUIRED = [str(requirement) for requirement in parse_requirements(requirements_txt)]


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "../README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = "Build and publish the package."
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print("\033[1m{0}\033[0m".format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status("Removing previous builds…")
            rmtree(os.path.join(here, "dist"))
        except OSError:
            pass

        self.status("Building Source and Wheel distribution…")
        os.system("{0} setup.py sdist bdist_wheel".format(sys.executable))

        self.status("Uploading the package to PyPI via Twine…")
        os.system("twine upload dist/*")

        print("🤖 Uploaded ...")
        sys.exit()


setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    packages=["python_rucaptcha"],
    install_requires=REQUIRED,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author_email=EMAIL,
    package_dir={"python-rucaptcha": "python_rucaptcha"},
    include_package_data=True,
    py_modules=["python_rucaptcha"],
    url=URL,
    license="AGPL-3.0",
    keywords="""
                captcha 
				rucaptcha
				2captcha
				recaptcha
				geetest
				hcaptcha
				capypuzzle
				tiktok
				rotatecaptcha
				funcaptcha
				keycaptcha
				python3
				recaptcha
				captcha
				security
				tiktok
				python-library
				python-rucaptcha
				rucaptcha-client
               """,
    python_requires=REQUIRES_PYTHON,
    zip_safe=False,
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Development Status :: 5 - Production/Stable",
        "Framework :: AsyncIO",
        "Operating System :: Unix",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
    ],
    # Build - `python setup.py bdist_wheel`
    # Upload package: `python3 setup.py upload`
    cmdclass={"upload": UploadCommand},
)
print("🤖 Success install ...")



from tenacity import AsyncRetrying, wait_fixed, stop_after_attempt
from requests.adapters import Retry

RETRIES = Retry(total=5, backoff_factor=0.5)
ASYNC_RETRIES = AsyncRetrying(wait=wait_fixed(5), stop=stop_after_attempt(5), reraise=True)
# Application key
APP_KEY = "1899"


# Connection retry generator
def connect_generator():
    for i in range(5):
        yield i



from enum import Enum
from typing import List


class MyEnum(Enum):
    @classmethod
    def list(cls) -> List[Enum]:
        return list(map(lambda c: c, cls))

    @classmethod
    def list_values(cls) -> List[str]:
        return list(map(lambda c: c.value, cls))

    @classmethod
    def list_names(cls) -> List[str]:
        return list(map(lambda c: c.name, cls))


class ServicesEnm(MyEnum):
    TWOCAPTCHA = "2captcha"
    RUCAPTCHA = "rucaptcha"


class SaveFormatsEnm(MyEnum):
    TEMP = "temp"
    CONST = "const"


class GeetestEnm(MyEnum):
    GEETEST = "geetest"
    GEETEST_V4 = "geetest_v4"


class ImageCaptchaEnm(MyEnum):
    BASE64 = "base64"


class CapyPuzzleEnm(MyEnum):
    CAPY = "capy"


class FunCaptchaEnm(MyEnum):
    FUNCAPTCHA = "funcaptcha"


class ReCaptchaEnm(MyEnum):
    USER_RECAPTCHA = "userrecaptcha"


class LeminCroppedCaptchaEnm(MyEnum):
    LEMIN = "lemin"


class HCaptchaEnm(MyEnum):
    HCAPTCHA = "hcaptcha"


class KeyCaptchaEnm(MyEnum):
    KEYCAPTCHA = "keycaptcha"


class RotateCaptchaEnm(MyEnum):
    ROTATECAPTCHA = "rotatecaptcha"


class TikTokCaptchaEnm(MyEnum):
    TIKTOK = "tiktok"


class CaptchaControlEnm(MyEnum):
    # default
    GET = "get"
    # https://rucaptcha.com/api-rucaptcha#manage_pingback
    ADD_PINGBACK = "add_pingback"
    GET_PINGBACK = "get_pingback"
    DEL_PINGBACK = "del_pingback"

    # https://rucaptcha.com/api-rucaptcha#additional
    GETBALANCE = "getbalance"
    GET2 = "get2"

    # https://rucaptcha.com/api-rucaptcha#complain
    REPORTGOOD = "reportgood"
    REPORTBAD = "reportbad"


class YandexSmartCaptchaEnm(MyEnum):
    YANDEX = "yandex"



from uuid import uuid4
from typing import Union

from pydantic import Field, BaseModel, validator, root_validator

from . import enums
from .config import APP_KEY

"""
Socket API Serializers
"""


class CaptchaOptionsSocketSer(BaseModel):
    phrase: bool = False
    caseSensitive: bool = False
    numeric: int = 0
    calc: bool = False
    minLen: int = 0
    maxLen: int = 0
    lang: str = ""
    hintText: str = ""
    hintImg: str = ""
    softId: str = "1899"

    @validator("numeric")
    def numeric_check(cls, value):
        if value not in range(1, 5):
            raise ValueError("Invalid `numeric` param value")
        return value

    @validator("minLen", "maxLen")
    def len_check(cls, value):
        if value not in range(0, 21):
            raise ValueError("Invalid `minLen \ maxLen` param value")
        return value

    @validator("hintText")
    def hint_text_check(cls, value):
        if len(value) > 140:
            raise ValueError("Invalid `hintText` param value")
        return value

    @validator("softId")
    def soft_id_set(cls, value):
        value.update({"softId": "1899"})
        return value


class NormalCaptchaSocketSer(BaseModel):
    method: str = "normal"
    requestId: str = Field(default_factory=lambda: str(uuid4()))
    body: str = str()
    options: "CaptchaOptionsSocketSer" = CaptchaOptionsSocketSer()


class TextCaptchaSocketSer(BaseModel):
    method: str = "text"
    requestId: str = Field(default_factory=lambda: str(uuid4()))
    body: str = str()
    options: "CaptchaOptionsSocketSer" = CaptchaOptionsSocketSer()


class ControlCaptchaSocketSer(BaseModel):
    method: str
    requestId: str = Field(default_factory=lambda: str(uuid4()))
    success: str = None
    captchaId: int = None


class SocketResponse(BaseModel):
    method: str = None
    success: bool = None
    code: str = None
    # captcha task ID at RuCaptcha service
    captchaId: int = None
    # manually generated requestID
    requestId: str = Field(default_factory=lambda: str(uuid4()))
    error: str = None
    # specific fields for balance request response
    balance: float = None
    valute: str = None


class SockAuthSer(BaseModel):
    method: str = "auth"
    requestId: str = Field(default_factory=lambda: str(uuid4()))
    key: str
    options: dict


"""
HTTP API Serializers
"""


class PostRequestSer(BaseModel):
    key: str
    method: str
    soft_id: str = APP_KEY
    field_json: int = Field(1, alias="json")


class GetRequestSer(BaseModel):
    key: str
    action: str = "get"
    field_json: int = Field(1, alias="json")

    # Control keys
    ids: str = None
    id: str = None


class CaptchaOptionsSer(BaseModel):
    rucaptcha_key: str
    method: str
    action: str
    sleep_time: int = 10
    service_type: str = enums.ServicesEnm.TWOCAPTCHA.value

    # CaptchaImage
    # save_format: str = enums.SaveFormatsEnm.TEMP.value
    # img_clearing: bool = True
    # img_path: str = "PythonRuCaptchaImages"

    url_request: str = ""
    url_response: str = ""

    @validator("rucaptcha_key")
    def len_check(cls, value):
        if len(value) != 32:
            raise ValueError("ERROR_WRONG_USER_KEY")
        return value

    @validator("sleep_time")
    def sleep_time_check(cls, value):
        if value < 5:
            raise ValueError("Too little sleep time")
        return value

    """@validator("save_format")
    def save_format_check(cls, value):
        if value not in enums.SaveFormatsEnm.list_values():
            raise ValueError(
                f"Invalid `save_format`, valid params - {','.join(enums.SaveFormatsEnm.list_values())}, send - {value}"
            )
        return value"""

    @validator("service_type")
    def service_type_check(cls, value):
        if value not in enums.ServicesEnm.list_values():
            raise ValueError(
                f"Invalid `service_type`, valid params - {','.join(enums.ServicesEnm.list_values())}, send - {value}"
            )
        return value

    @root_validator
    def urls_set(cls, values):
        service_type = values.get("service_type")
        values.update(
            {"url_request": f"http://{service_type}.com/in.php", "url_response": f"http://{service_type}.com/res.php"}
        )
        return values


"""
HTTP API Response
"""


class ServicePostResponseSer(BaseModel):
    status: int
    request: str


class ServiceGetResponseSer(BaseModel):
    status: int
    request: Union[str, dict]

    # ReCaptcha V3 params
    user_check: str = ""
    user_score: str = ""


class ResponseSer(BaseModel):
    serverAnswer: dict = {}
    captchaSolve: dict = {}
    taskId: int = None
    error: bool = False
    errorBody: str = None

    @validator("taskId", pre=True, always=True)
    def dt_check(cls, value):
        if value:
            value = int(value)
        return value
